from __future__ import annotations
import typing
from timeit import default_timer as timer

import numpy as np
from sklearn.neighbors import NearestNeighbors

from guipilot.matcher import WidgetMatcher, Pair, Score

if typing.TYPE_CHECKING:
    from guipilot.entities import Screen
    

class GVT(WidgetMatcher):
    def __init__(self, threshold) -> None:
        super().__init__()
        self.threshold = threshold

    def match(self, screen_i: Screen, screen_j: Screen) -> tuple[list[Pair], list[Score], float]:
        start_time = timer()
        widget_keys_i, widget_keys_j = list(screen_i.widgets.keys()), list(screen_j.widgets.keys())
        points_i = np.array([list(self._norm_xywh(screen_i, widget)) for widget in screen_i.widgets.values()])
        points_j = np.array([list(self._norm_xywh(screen_j, widget)) for widget in screen_j.widgets.values()])

        knn = NearestNeighbors(n_neighbors=1, metric="manhattan")
        knn.fit(points_j)
        distances, indices = knn.kneighbors(points_i)
        sorted_distances_indices = sorted(
            enumerate(zip(distances, indices)),
            key=lambda x: x[1][0][0]  # Sort by the distance
        )

        paired_ids = set()
        pairs, scores = [], []
        for i, (distance, index) in sorted_distances_indices:
            if distance[0] <= self.threshold:
                widget_i = widget_keys_i[i]
                widget_j = widget_keys_j[int(index[0])]

                # Skip if either widget has already been paired
                if widget_i in paired_ids or widget_j in paired_ids:
                    continue

                # Add pair to the list
                pairs.append((widget_i, widget_j))
                scores.append(1 / distance[0])

                # Mark widgets as paired
                paired_ids.add(widget_i)
                paired_ids.add(widget_j)

        time = (timer() - start_time) * 1000
        return pairs, scores, int(time)


from __future__ import annotations
import typing
from enum import Enum
from dataclasses import dataclass, field

if typing.TYPE_CHECKING:
    from .constants import Bbox


class WidgetType(Enum):
    TEXT_BUTTON = "textbutton"
    ICON_BUTTON = "iconbutton"
    COMBINED_BUTTON = "combinedbutton"
    INPUT_BOX = "inputbox"
    TEXT_VIEW = "textview"
    IMAGE_VIEW = "imageview"
    CHART = "chart"
    SLIDER = "slider"


@dataclass
class Widget:
    type: WidgetType
    bbox: Bbox
    texts: list[str] = field(default_factory=list)
    text_bboxes: list[Bbox] = field(default_factory=list)

    @property
    def width(self) -> float:
        return self.bbox.xmax - self.bbox.xmin
    
    @property
    def height(self) -> float:
        return self.bbox.ymax - self.bbox.ymin
    
    @property
    def area(self) -> float:
        return self.width * self.height
    
    @property
    def center(self) -> tuple[float, float]:
        cx = (self.bbox.xmin + self.bbox.xmax) / 2
        cy = (self.bbox.ymin + self.bbox.ymax) / 2
        return cx, cy


from __future__ import annotations
import typing
from dataclasses import dataclass, field

import cv2
import numpy as np

from .constants import Bbox
from .widget import Widget, WidgetType
from guipilot.models import OCR, Detector

if typing.TYPE_CHECKING:
    from .screen import Screen
    from guipilot.checker import ScreenChecker
    from guipilot.matcher import WidgetMatcher


ocr = OCR(service_url="http://localhost:5000/detect")
detector = Detector(service_url="http://localhost:6000/detect")


@dataclass
class Screen:
    image: np.ndarray
    widgets: dict[int, Widget] = field(default_factory=dict)
    
    def detect(self) -> None:
        """
        Use object detector to extract widgets from screen
        Updates widgets list to the detection results
        """
        def _to_bbox(points: np.ndarray) -> Bbox:
            xmin, ymin, xmax, ymax = points
            return Bbox(int(xmin), int(ymin), int(xmax), int(ymax))
        
        bboxes, widget_types = detector(self.image)
        self.widgets = {
            i: Widget(
                type=WidgetType(widget_type),
                bbox=_to_bbox(bbox)
            )
            for i, (bbox, widget_type) in enumerate(zip(bboxes, widget_types))
        }

        assert len(self.widgets) == len(bboxes) == len(widget_types)

    def ocr(self) -> None:
        """
        Use OCR to extract text from screen and assign to widgets
        """
        image = np.array(self.image)
        image = cv2.cvtColor(image, cv2.COLOR_BGR2RGB)
        widgets = self.widgets.values()
        for widget in widgets:
            xmin, ymin, xmax, ymax = widget.bbox
            widget_image = image[ymin:ymax, xmin:xmax].copy()
            try:
                widget.texts, widget.text_bboxes = ocr(widget_image)
            except Exception as e:
                print(e)
                print(self.image.shape)
                print(widget.bbox)

    def check(self, target: Screen, matcher: WidgetMatcher, checker: ScreenChecker) -> tuple[set, float]:
        """Check for screen inconsistency

        Args:
            target: the screen to check against
            matcher: algorithm to match same widgets as pairs on both screens
            checker: algorithm to check consistency of widget pairs

        Returns:
            see ScreenChecker
        """
        pairs = matcher.match(self, target)
        inconsistencies, time_taken = checker.check(self, target, pairs)
        return inconsistencies, time_taken


from __future__ import annotations
import typing
from timeit import default_timer as timer

import numpy as np

from guipilot.matcher import WidgetMatcher, Pair, Score

if typing.TYPE_CHECKING:
    from guipilot.entities import Screen, Widget


class GUIPilotV2(WidgetMatcher):
    def __init__(self, s1: int = 100, s2: int = 1) -> None:
        """
        Params
            s1: scaling factor for distance score i.e. s1 * (abs(xi - xj) + abs(yi - yj))
            s2: scaling factor for shape score i.e. s2 * (abs(wi - wj) + abs(hi - hj))
        """
        self.s1, self.s2 = s1, s2
        super().__init__()

    def match(self, screen_i: Screen, screen_j: Screen) -> tuple[list[Pair], list[Score], float]:
        start_time = timer()
        widget_keys_i, widget_keys_j = list(screen_i.widgets.keys()), list(screen_j.widgets.keys())
        scores = self._calculate_match_scores(screen_i, screen_j)
        path = self._find_longest_matching_subsequence(scores)
        pairs = [(widget_keys_i[x], widget_keys_j[y]) for x, y in path]
        scores = [scores[x, y] for (x, y) in path]
        time = (timer() - start_time) * 1000
        return pairs, scores, int(time)

    def _calculate_match_scores(self, screen_i: Screen, screen_j: Screen) -> np.ndarray:
        def get_distance_score(si: Screen, sj: Screen, i: Widget, j: Widget) -> float:
            """Calculates Manhattan distance between normalized bboxes of widgets.
            """
            si_h, si_w, _ = si.image.shape
            sj_h, sj_w, _ = sj.image.shape

            xi, yi, wi, hi = i.bbox.xmin / si_w, i.bbox.ymin / si_h, i.width / si_w, i.height / si_h
            xj, yj, wj, hj = j.bbox.xmin / sj_w, j.bbox.ymin / sj_h, j.width / sj_w, j.height / sj_h
            
            score = self.s1 * (abs(xi - xj) + abs(yi - yj)) + self.s2 * (abs(wi - wj) + abs(hi - hj))
            return min(1 / score, 1) if score else 1

        def get_area_score(i: Widget, j: Widget) -> float:
            """Calculates the ratio of widget areas.
            """
            areas = [i.area, j.area]
            return min(areas) / max(areas)

        def get_shape_score(i: Widget, j: Widget) -> float:
            """Calculate the ratio of aspect ratios of widgets.
            """
            aspect_ratios = [i.width / i.height, j.width / j.height]
            return min(aspect_ratios) / max(aspect_ratios)

        def get_type_score(i: Widget, j: Widget) -> float:
            """Compare the type compatibility of widgets.
            """
            if i.type == j.type: return 1
            else: return 0.01
        
        widgets_i, widgets_j = screen_i.widgets.values(), screen_j.widgets.values()
        scores = np.zeros((len(widgets_i), len(widgets_j)))

        for i, widget_i in enumerate(widgets_i):
            for j, widget_j in enumerate(widgets_j):
                distance_score = get_distance_score(screen_i, screen_j, widget_i, widget_j)
                area_score = get_area_score(widget_i, widget_j)
                shape_score = get_shape_score(widget_i, widget_j)
                type_score = get_type_score(widget_i, widget_j)

                score = distance_score * area_score * shape_score * type_score
                scores[i, j] = max(score, 1e-8)

        return scores

    def _find_longest_matching_subsequence(self, D: np.ndarray) -> list[tuple]:
        m, n = D.shape
        if m == 0 or n == 0:
            return []
        dp = np.zeros((m + 1, n + 1))
        for i in range(1, m + 1):
            for j in range(1, n + 1):
                dp[i, j] = max(
                    dp[i - 1, j], dp[i, j - 1], dp[i - 1, j - 1] + D[i - 1, j - 1]
                )
        i, j = m, n
        sequence = []
        while i > 0 and j > 0:
            if dp[i, j] == dp[i - 1, j - 1] + D[i - 1, j - 1]:
                sequence.append((i - 1, j - 1))
                i, j = i - 1, j - 1
            elif dp[i, j] == dp[i - 1, j]:
                i -= 1
            else:
                j -= 1
        return sequence[::-1]


import io
import os
import base64
from abc import ABC, abstractmethod
from typing import Optional

import openai
from PIL import Image


class Agent(ABC):
    @abstractmethod
    def __call__(self, prompt: str, images: Optional[list[Image.Image]] = None) -> str:
        pass


class GPTAgent(Agent):
    def __init__(self, api_key: str, model: str = "gpt-4o") -> None:
        base_path = os.path.abspath(os.path.dirname(__file__))
        self.model = model
        self.client = openai.OpenAI(api_key=api_key)
        self.system_prompt = open(f"{base_path}/action_completion.system.prompt").read()
        self.history = [{"role": "system", "content": self.system_prompt}]

    def reset(self):
        self.history = []

    def __call__(self, prompt: str, images: Optional[list[Image.Image]] = None) -> str:
        if len(self.history) > 0:
            self.history.append({
                "role": "user", 
                "content": "Incorrect, are you sure it is the correct: 1) widget id, 2) action type 3) direction (i.e., swipe/scroll)? Do not use the same answer again."
            })

        user_prompt = [{"type": "text", "text": prompt}]

        for image in images:
            bytes = io.BytesIO()
            image.save(bytes, format="JPEG")
            image_b64 = base64.b64encode(bytes.getvalue()).decode('ascii')
            user_prompt.append({
                "type": "image_url",
                "image_url": {"url": f"data:image/jpeg;base64,{image_b64}"}
            })


        self.history.append({"role": "user", "content": user_prompt})
        response = self.client.chat.completions.create(
            model=self.model,
            messages=self.history,
            max_tokens=1024,
            temperature=0,
            top_p=1
        )

        content = response.choices[0].message.content
        self.history.append({"role": "assistant", "content": content})

        return content


import numpy as np
import matplotlib.pyplot as plt
from matplotlib.figure import Figure
from matplotlib.axes import Axes


def visualize_match_scores(
    scores: np.ndarray, 
    path: list[tuple], 
    widget_keys_i: list[int], 
    widget_keys_j: list[int]
) -> tuple[Figure, Axes]:
    
    fig, ax = plt.subplots(figsize=(12, 10))
    
    # Display the grid
    ax.imshow(scores, cmap='viridis', interpolation='nearest')

    # Highlight the path
    x_path, y_path = zip(*path)
    ax.plot(y_path, x_path, color='red', linewidth=2, marker='o')

    # Annotate the grid with values
    for (i, j), val in np.ndenumerate(scores):
        val = round(val, 2)
        ax.text(j, i, f'{val}', ha='center', va='center', color='white', fontsize=8)

    # Set x and y ticks using widget keys
    ax.set_xticks(np.arange(scores.shape[1]), [x for x in widget_keys_j])
    ax.set_xticklabels(widget_keys_j)
    ax.set_yticks(np.arange(scores.shape[0]), [y for y in widget_keys_i])
    ax.set_yticklabels(widget_keys_i)

    # Add color bar, labels, and title
    fig.colorbar(label='Score')
    ax.set_ylabel('original')
    ax.set_xlabel('mutated')
    ax.set_title('2D Grid of Scores with Highlighted Path')

    return fig, ax


from __future__ import annotations
import typing
from abc import ABC, abstractmethod
from timeit import default_timer as timer

import numpy as np

if typing.TYPE_CHECKING:
    from guipilot.entities import Widget, Screen


class ScreenChecker(ABC):
    def check(self, screen_i: Screen, screen_j: Screen, pairs: list[tuple[int, int]]) -> tuple[set, float]:
        """Checks for widget inconsistencies on two screens.

        Args:
            screen_i, screen_j: two screens containing a list of widgets to compare
            pairs: a list of tuples, where each tuple `(x, y)` represents a pair of matching 
            widget IDs. `x` is from `screen_i` and `y` is from `screen_j`.

        Returns:
            set: A set of tuples containing:
            - Inconsistent (i, j, type): widget ID pairs with bbox, text, or color inconsistencies
            - Missing (i, None): widget IDs in screen_i that are not paired
            - Excess (None, j): widget IDs in screen_j that are not paired

            float: Time taken (seconds) to check all widgets
        """
        unpaired_i = set([id for id in screen_i.widgets.keys()])
        unpaired_j = set([id for id in screen_j.widgets.keys()])

        start_time = timer()
        result = set()
        for pair in pairs:
            x, y = pair
            unpaired_i.discard(x)
            unpaired_j.discard(y)

            widget_i = screen_i.widgets[x]
            xmin, ymin, xmax, ymax = widget_i.bbox
            widget_image_i = screen_i.image[ymin:ymax, xmin:xmax]

            widget_j = screen_j.widgets[y]
            xmin, ymin, xmax, ymax = widget_j.bbox
            widget_image_j = screen_j.image[ymin:ymax, xmin:xmax]

            inconsistencies = self.check_widget_pair(widget_i, widget_j, widget_image_i, widget_image_j)
            result.update([(x, y, k) for k in inconsistencies])

        result.update([(id, None) for id in unpaired_i])
        result.update([(None, id) for id in unpaired_j])
        time = (timer() - start_time) * 1000
        return result, int(time)

    @abstractmethod
    def check_widget_pair(self, w1: Widget, w2: Widget, wi1: np.ndarray, wi2: np.ndarray) -> list[tuple]:
        """Check if a pair of widgets are consistent.

        Args:
            w1, w2: Widget pairs to check
            wi1, wi2: Widget images to check

        Returns:
            A list of tuples, see check() for explanation.
        """
        pass


from __future__ import annotations
import typing
import re
import cv2
import numpy as np
from difflib import SequenceMatcher

from .checker import ScreenChecker
from guipilot.entities import WidgetType, Inconsistency

if typing.TYPE_CHECKING:
    from guipilot.entities import Widget


class GUIPilot(ScreenChecker):
    def check_widget_pair(self, w1: Widget, w2: Widget, wi1: np.ndarray, wi2: np.ndarray) -> list[tuple]:
        def check_bbox_consistency(w1: Widget, w2: Widget) -> bool:
            """Check if both widgets have similar position, size, and shape on the screen
            """
            xa, ya = max(w1.bbox[0], w2.bbox[0]), max(w1.bbox[1], w2.bbox[1])
            xb, yb = min(w1.bbox[2], w2.bbox[2]), min(w1.bbox[3], w2.bbox[3])
            intersection = abs(max((xb - xa, 0)) * max((yb - ya), 0))
            boxa = abs((w1.bbox[2] - w1.bbox[0]) * (w1.bbox[3] - w1.bbox[1]))
            boxb = abs((w2.bbox[2] - w2.bbox[0]) * (w2.bbox[3] - w2.bbox[1]))
            iou = intersection / (boxa + boxb - intersection)
            return iou > 0.9

        def check_text_consistency(w1: Widget, w2: Widget) -> bool:
            """Check if the text on both widgets are similar
            """
            has_text = {WidgetType.TEXT_VIEW, WidgetType.TEXT_BUTTON, WidgetType.COMBINED_BUTTON, WidgetType.INPUT_BOX}
            if w1.type not in has_text or w2.type not in has_text: return True

            for t1, t2 in zip(w1.texts, w2.texts):
                t1 = re.sub(r'[^a-zA-Z0-9]', '', t1)
                t2 = re.sub(r'[^a-zA-Z0-9]', '', t2)
                if SequenceMatcher(None, t1.lower(), t2.lower()).quick_ratio() < 0.95: return False
            
            return True

        def check_color_consistency(wi1: np.ndarray, wi2: np.ndarray) -> bool:
            """Check if the color distribution on both widgets are similar
            """
            # normalized 3D color histogram, 8 bins per channel
            hist1 = cv2.calcHist([wi1], [0, 1, 2], None, [8, 8, 8], [0, 250, 0, 250, 0, 250])
            hist1 = cv2.normalize(hist1, hist1).flatten()
            
            hist2 = cv2.calcHist([wi2], [0, 1, 2], None, [8, 8, 8], [0, 250, 0, 250, 0, 250])
            hist2 = cv2.normalize(hist2, hist2).flatten()

            score = cv2.compareHist(hist1, hist2, cv2.HISTCMP_KL_DIV)
            return score < 8
        
        diff = set()
        if not check_bbox_consistency(w1, w2): diff.add(Inconsistency.BBOX)
        if not check_text_consistency(w1, w2): diff.add(Inconsistency.TEXT)
        if Inconsistency.TEXT not in diff:
            if not check_color_consistency(wi1, wi2): diff.add(Inconsistency.COLOR)
            
        return list(diff)


from __future__ import annotations
import typing
from timeit import default_timer as timer

if typing.TYPE_CHECKING:
    from .screen import Screen
    from guipilot.checker import ScreenChecker
    from guipilot.matcher import WidgetMatcher


class Process(object):
    def __init__(self) -> None:
        self.screens: list[Screen] = []
    
    def add(self, screen: Screen) -> None:
        """Add a screen to the end of process
        """
        self.screens.append(screen)
    
    def check(self, target: Screen, matcher: WidgetMatcher, checker: ScreenChecker, process_path, i) -> tuple[bool, float]:
        """Check for process inconsistency on the current screen

        The current screen is the most recent screen the process is on. This method compares the 
        target screen against all screens in the process. If the target screen has the fewest 
        inconsistencies with the current screen, the process is considered consistent so far.

        Args:
            target: The screen to check against
            matcher: An algorithm that pairs matching widgets on both screens.
            checker: An algorithm that checks the consistency of paired widgets.

        Returns:
            bool: True if the process is consistent, otherwise False.
        """
        start_time = timer()
        screen = self.screens[-1]
        pairs, scores = matcher.match(screen, target)
        inconsistencies, _ = checker.check(screen, target, pairs)

        unpaired_screen = set([x[0] for x in inconsistencies if x[1] is None])
        unpaired_screen = len(unpaired_screen) / len(screen.widgets)

        unpaired_target = set([x[1] for x in inconsistencies if x[0] is None])
        unpaired_target = len(unpaired_target) / len(target.widgets)

        matching_score = sum(scores) / len(target.widgets)
        
        a = matching_score
        b = (matching_score - unpaired_screen)
        c = (matching_score - unpaired_target)

        return (a, b, c), timer() - start_time


from __future__ import annotations
import typing
from typing import TypeAlias
from abc import ABC, abstractmethod

from guipilot.entities import Bbox

if typing.TYPE_CHECKING:
    from guipilot.entities import Widget, Screen


Pair: TypeAlias = tuple[int, int]
Score: TypeAlias = float
    

class WidgetMatcher(ABC):
    @abstractmethod
    def match(self, screen_i: Screen, screen_j: Screen) -> tuple[list[Pair], list[Score], float]:
        """Match widgets between two screens

        Args:
            screen_i, screen_j: two screens containing a list of widgets to match

        Returns:
            A list of tuples, where each tuple `(x, y)` represents a pair of matching 
            widget IDs. `x` is from `screen_i` and `y` is from `screen_j`.
        """
        pass
    
    def _norm_xywh(self, screen: Screen, widget: Widget) -> tuple[float, float, float, float]:
        """Calculate the normalized bounding box of a widget
        """
        screen_height, screen_width, _ = screen.image.shape
        xmin, ymin, xmax, ymax = widget.bbox
        xmin, xmax = xmin / screen_width, xmax / screen_width
        ymin, ymax = ymin / screen_height, ymax / screen_height
        assert screen_height > screen_width
        assert 0 <= xmin <= 1 and 0 <= xmax <= 1
        assert 0 <= ymin <= 1 and 0 <= ymax <= 1
        assert xmin <= xmax and ymin <= ymax
        return xmin, ymin, xmax - xmin, ymax - ymin


from __future__ import annotations
import typing
import re
import math
from difflib import SequenceMatcher

import cv2
import numpy as np
from PIL import Image

from .checker import ScreenChecker
from guipilot.entities import WidgetType, Inconsistency

if typing.TYPE_CHECKING:
    from guipilot.entities import Widget


class GVT(ScreenChecker):
    def check_widget_pair(self, w1: Widget, w2: Widget, wi1: np.ndarray, wi2: np.ndarray) -> list[tuple]:
        def get_quantized_colors(image: np.ndarray, k=3) -> list[tuple[int, int, int]]:
            """Extract k main colors from the image 
            """
            image = cv2.cvtColor(image, cv2.COLOR_BGR2RGB)
            image: Image.Image = Image.fromarray(image)
            image = image.quantize(colors=3, method=Image.Quantize.MEDIANCUT, dither=Image.NONE, kmeans=0)
            palette = image.getpalette()
            colors_list: list[int] = [palette[i : i + 3] for i in range(0, k * 3, 3)]
            return [[255, 255, 255] if not color else color for color in colors_list]

        def get_color_distance(color1: tuple[int, int, int], color2: tuple[int, int, int]) -> float:
            """Calculates redmean color distance, weight adjusted for human perception
            """
            r1, g1, b1 = color1
            r2, g2, b2 = color2
            max_dist = 764.8339663572415
            mean_r = (r1 + r2) / 2
            delta_r, delta_g, delta_b = r1 - r2, g1 - g2, b1 - b2
            weight_r, weight_g, weight_b = 2 + mean_r / 256, 4, 2 + (255 - mean_r) / 256
            dist = math.sqrt(weight_r * delta_r**2 + weight_g * delta_g**2 + weight_b * delta_b**2)
            return dist / max_dist
        
        def check_bbox_consistency(w1: Widget, w2: Widget) -> bool:
            """Check if both widgets have similar position, size, and shape on the screen
            """
            xa, ya = max(w1.bbox[0], w2.bbox[0]), max(w1.bbox[1], w2.bbox[1])
            xb, yb = min(w1.bbox[2], w2.bbox[2]), min(w1.bbox[3], w2.bbox[3])
            intersection = abs(max((xb - xa, 0)) * max((yb - ya), 0))
            boxa = abs((w1.bbox[2] - w1.bbox[0]) * (w1.bbox[3] - w1.bbox[1]))
            boxb = abs((w2.bbox[2] - w2.bbox[0]) * (w2.bbox[3] - w2.bbox[1]))
            iou = intersection / (boxa + boxb - intersection)
            return iou > 0.9
        
        def check_color_consistency(wi1: np.ndarray, wi2: np.ndarray) -> bool:
            """Perform color quantization to get color histograms, compare redmean distance between color pairs from histograms
            """
            colors1 = get_quantized_colors(wi1, 3)
            colors2 = get_quantized_colors(wi2, 3)
            return all([get_color_distance(color1, color2) <= 0.01 for color1, color2 in zip(colors1, colors2)])

        def check_pid_consistency(wi1: np.ndarray, wi2: np.ndarray) -> bool:
            """Perform binary thresholding and use perceptual image differencing on binarized images
            """
            h1, w1, _ = wi1.shape
            h2, w2, _ = wi2.shape
            h3, w3 = max(h1, h2), max(w1, w2)
            wi1 = cv2.cvtColor(wi1, cv2.COLOR_BGR2GRAY)
            wi1 = cv2.resize(wi1, (w3, h3), interpolation=cv2.INTER_AREA)
            wi2 = cv2.cvtColor(wi2, cv2.COLOR_BGR2GRAY)
            wi2 = cv2.resize(wi2, (w3, h3), interpolation=cv2.INTER_AREA)
            absdiff = cv2.absdiff(wi1, wi2)
            _, thresholded = cv2.threshold(absdiff, int(0.1 * 255), 255, cv2.THRESH_BINARY)
            diff_ratio = np.count_nonzero(thresholded) / (h3 * w3)
            return diff_ratio <= 0.2 
        
        def check_text_consistency(w1: Widget, w2: Widget) -> bool:
            """Check if the text on both widgets are similar
            """
            has_text = {WidgetType.TEXT_VIEW, WidgetType.TEXT_BUTTON, WidgetType.COMBINED_BUTTON, WidgetType.INPUT_BOX}
            if w1.type not in has_text or w2.type not in has_text: return True

            for t1, t2 in zip(w1.texts, w2.texts):
                t1 = re.sub(r'[^a-zA-Z0-9]', '', t1)
                t2 = re.sub(r'[^a-zA-Z0-9]', '', t2)
                if SequenceMatcher(None, t1.lower(), t2.lower()).quick_ratio() < 0.95: return False
            
            return True

        diff = set()
        if not check_bbox_consistency(w1, w2): diff.add(Inconsistency.BBOX)
        if not check_text_consistency(w1, w2): diff.add(Inconsistency.TEXT)
        if not check_color_consistency(wi1, wi2): diff.add(Inconsistency.COLOR)
        if not check_pid_consistency(wi1, wi2): diff.add(Inconsistency.COLOR)
            
        return list(diff)